function [X,y] = generate_ripley (Nn, Np, k, sigma)
% function [X,y] = generate_ripley (Nn, Np, k, sigma)
%
% Generate Ripley dataset
%
% Inputs:
%   - Nn, Np (integer): Number of samples of negative and positive
%     class. Must be divisible by two.
%   - k (integer): number of features.
%   - sigma (scalar): standard deviation of gaussian distributions
% Outputs:
%   - X ((Nn+Np) x k): matrix of features
%   - y ((Nn+Np)x1): labels (-1 or 1)
%
% J. Spilka, P. Abry, R. Leonarduzzi
% 2018

if rem (Np, 2) ~= 0
    error('generate_ripley: Set the number of sample to be divisible by 2');
end

if rem (Nn, 2) ~= 0
    error('generate_ripley: Set the number of sample to be divisible by 2');
end

if rem(k,2) ~= 0
    error('generate_ripley: Number of features must be divisible by 2');
end

if ~exist('sigma','var')
    sigma = 0.04;
end

% Create k / 2 realizations of bidimensional Ripley features
X = NaN (Nn + Np, k);
for i = 1 : k / 2
    % Features for class 1
    X1_neg = mvnrnd ([ 0.4, 0.7], [sigma, sigma], Nn / 2);
    X2_neg = mvnrnd ([-0.3, 0.7], [sigma, sigma], Nn / 2);

    % Features for class 2
    X1_pos = mvnrnd ([-0.7, 0.3], [sigma, sigma], Np / 2);
    X2_pos = mvnrnd ([ 0.3, 0.3], [sigma, sigma], Np / 2);

    % Concatenate features
    X(:, 2*(i-1)+(1:2)) = [X1_neg ; X2_neg ; X1_pos ; X2_pos];
end

y = [-ones(Nn,1); ones(Np,1)];